package sofie.maxsat;

import java.io.File;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import javatools.administrative.Announce;

/**
 * Class HypothesisGenerator
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *
 * Runs the FMSStar Algorithm
 *  
 * @author Fabian M. Suchanek
 */
public class FMSStar {

  /** Sets all hypotheses, TRUE if hypotheses were generated */
  public static boolean run(int from, int to, File outputFile) throws SQLException {
    Config.setPatternOccurences(from, to);
    // Generate the hypotheses
    if(!HypothesisStore.generateHypotheses()) return(false);
    // Figure out the priorities of the hypotheses
    Announce.progressStart("Determining the priorities", HypothesisStore.numHypotheses());
    int counter = 0;
    SortedSet<Hypothesis> possiblyDetermined = new TreeSet<Hypothesis>();
    for (Hypothesis hyp : HypothesisStore.hypotheses()) {
      Announce.progressAt(counter++);
      Measures measures = HypothesisStore.computeMeasures(hyp);
      if (measures.isDetermined()) {
        hyp.setValue(measures.truthValue());
        possiblyDetermined.addAll(measures.relatedHypotheses());
      }
    }
    Announce.progressDone();

    // Run the algorithm
    Announce.progressStart("Running FMSStar algorithm", HypothesisStore.numHypotheses());
    int numSetHypotheses = 0;
    while (true) {
      Announce.progressAt(numSetHypotheses);
      // Check all hypotheses that are possibly determined, see whether they are really determined
      if (possiblyDetermined.size() != 0) {
        Hypothesis hyp = possiblyDetermined.first();
        possiblyDetermined.remove(hyp);
        if (hyp.value() != -1) continue;
        Measures measures = HypothesisStore.computeMeasures(hyp);
        if (measures.isDetermined()) {
          hyp.setValue(measures.truthValue());
          numSetHypotheses++;
          possiblyDetermined.addAll(measures.relatedHypotheses());
          continue;
        }
      }
      // Set the best hypothesis
      Hypothesis hyp = HypothesisStore.getHypothesisToSet();
      if (hyp == null) break;
      Measures measures = HypothesisStore.computeMeasures(hyp);
      hyp.setValue(measures.truthValue());
      numSetHypotheses++;
      possiblyDetermined.addAll(measures.relatedHypotheses());
    }
    Announce.progressDone(); 
    saveTo(outputFile);
    return(true);
  }

  /** Constant for "save to YAGO instead of a file"*/
  protected static final File YAGO = new File(".");

  /** Saves the output  */
  public static void saveTo(File output) {
    if (output == YAGO) {
      try {
        HypothesisStore.saveToYAGO();
      } catch (Exception e) {
        e.printStackTrace();
        output = null;
      }
    }
    
    if (output != null && output != YAGO) {
      try {
        HypothesisStore.saveToFile(output);
      } catch (Exception e) {
        e.printStackTrace();
        output = null;
      }
    }
    
    if (output == null) {
      System.out.println("\n----------------- Hypotheses");
      System.out.println(HypothesisStore.hypotheses());
    }
  }

  // -------------------------------------------------------------------------
  //                       Main
  // -------------------------------------------------------------------------

  /** Recomputes all facts*/
  public static void main(String[] args) throws Exception {
    // Help
    if (args.length > 0 && Announce.isHelp(args[0]) || args.length>2) {
      Announce.help("     sofie.maxsat.FMSStar [<ini-file>] [<output-file>]\n", "Runs the FMSStar Algorithm.",
          "<ini-file>    an initialization file that defines the database and other parameters.", 
          "              Defaults to sofie.ini .",
          "<output-file> a file where the results are stored.", 
          "              Defaults to the screen. If results of one batch shall improve",                    
          "              the results of a previous batch, use <output-file>=YAGO.",
          "              This will write the results into the YAGO database."
      );
    }
    
    // Parse the arguments
    List<String> argv=new ArrayList<String>(Arrays.asList(args));
    File iniFile = new File("sofie.ini");
    if(argv.size()>0 && argv.get(0).endsWith(".ini")) {
      iniFile=new File(argv.get(0));
      argv.remove(0);
    }
    File outputFile=null;
    if(argv.size()>0) {
      if(argv.get(0).equalsIgnoreCase("yago")) {
        outputFile=YAGO;
      }
      else {
        outputFile=new File(argv.get(0));
        outputFile.delete();
      }
    }
    
    // Initialize
    Config.init(iniFile);
    // Run the batches
    for(int start=0;;start+=Config.BATCHSIZE) {
      if(!run(start, start+Config.BATCHSIZE-1, outputFile)) break;          
    }    
  }
}
