package sofie.maxsat;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import javatools.administrative.Announce;
import javatools.database.Database;
import javatools.database.ResultIterator;
import javatools.filehandlers.CSVFile;

/**
 * Class HypothesisStore
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *      
 * This class stores all hypotheses.
 * 
 * @author Fabian M. Suchanek
 */
public class HypothesisStore {

  // -------------------------------------------------------------------------
  //                       Hyopothesis set
  // -------------------------------------------------------------------------

  /** Holds the hypotheses*/
  protected static SortedSet<Hypothesis> hypotheses = new TreeSet<Hypothesis>();

  /** A flag that stores whether we are in generation mode*/
  protected static boolean isGenerating=false;
  
  /** Returns the set of hypotheses*/
  public static SortedSet<Hypothesis> hypotheses() {
    return (hypotheses);
  }

  /** Returns the hypothesis that is not set and has highest priority (or null). Runs linear time.*/
  public static Hypothesis getHypothesisToSet() {
    if (hypotheses.size() == 0) return (null);
    Hypothesis best = null;
    for (Hypothesis hyp : hypotheses) {
      if (hyp.value() != -1) continue;
      if (best == null) {
        best = hyp;
        continue;
      }
      if (best.priority() < hyp.priority()) best = hyp;
    }
    return (best);
  }

  /** Returns number of hypotheses*/
  public static int numHypotheses() {
    return (hypotheses.size());
  }

  /** Adds a set of hypotheses to the hypothesis store*/
  public static void add(Collection<Hypothesis> hyps) {
    if(!isGenerating) return;
    for(Hypothesis hyp : hyps) {
      if(hyp.value()==-1) {
        if(hyp.trace()) Announce.message("Adding ",hyp,"to the hypothesis store");
        hypotheses.add(hyp);
      }
    }
  }
  
  /** Returns hypotheses that match a literal and that are unsatisfied or unknown. Runs log+linear time*/
  public static List<Hypothesis> matchingUnhappyHypotheses(Literal lit) throws SQLException {
    if (lit.trace()) Announce.message("Finding matching hypotheses for", lit);

    // If the literall is fully instantiated, find it and return it if unhappy
    if (!lit.hasVariable()) {
        Hypothesis newHyp;
        if (hypotheses.contains(lit)) {
          newHyp = hypotheses.tailSet(lit).first();
          if (lit.trace()) Announce.message("Found with value",newHyp.value());
          if (lit.isUnhappyWith(newHyp)) return (Arrays.asList(newHyp));
          return (Arrays.asList());
        }
        if (lit.hasInstances()) {
          return (Arrays.asList()); // It has at most 1 instance
        }
        newHyp = new Hypothesis(-1, lit.components());
        if(!isGenerating) newHyp.setValue(false);
        if (lit.trace()) Announce.message("Inventing", newHyp);
        return (Arrays.asList(newHyp));
    }

    // Find the literal that ends the search
    Literal end = new Literal(true, new ArrayList<String>(lit.components()));
    boolean foundVar = false;
    for (int i = 0; i < end.numComponents(); i++) {
      if (Literal.isVariable(end.component(i))) foundVar = true;
      if (foundVar) end.components().set(i, "? " + i);
    }
    // Collect the hypotheses in a list instead of an iterator to avoid
    // concurrent modification problems
    List<Hypothesis> result = new ArrayList<Hypothesis>();    
    for (Hypothesis h : hypotheses.tailSet(lit)) {
      if (!end.matches(h)) break;
      if (!lit.matches(h)) continue;
      result.add(h);
    }
    return (result);
  }

  // -------------------------------------------------------------------------
  //                       Saving
  // -------------------------------------------------------------------------

  /** Writes the true hypotheses to YAGO, excluding hypotheses with polysems*/
  public static void saveToYAGO() throws Exception {
    long id;
    Announce.doing("Retrieving latest free id from YAGO");
    try {
      id = Config.database.queryValue("SELECT MAX(id) FROM facts", ResultIterator.LongWrapper);
    } catch (Exception ex) {
      Announce.failed();
      throw ex;
    }
    Announce.done();
    saveToYAGO(id + 1);
  }

  /** Writes the true hypotheses to YAGO, excluding hypotheses with polysems*/
  public static void saveToYAGO(long startId) throws Exception {
    Announce.progressStart("Saving hypotheses to YAGO at id "+startId, hypotheses.size());
    try {
      Database.Inserter hypBulki = Config.database.newInserter("facts");
      for (Hypothesis hyp : hypotheses) {
        Announce.progressStep();
        if (hyp.value() == 1 && !hyp.containsPolysem()) hyp.saveTo(startId++, hypBulki);
      }
      hypBulki.close();
    } catch (Exception e) {
      Announce.failed();
      throw e;
    }
    Announce.progressDone();
  }

  /** Writes the hypotheses to a file */
  public static void saveToFile(File file) throws IOException {
    try {
      Announce.progressStart("Appending hypotheses to " + file.getName(), hypotheses.size());
      int counter = 0;
      CSVFile out = new CSVFile(file, true, "id","relation","arg1","arg2");
      for (Hypothesis hyp : hypotheses) {
        if (hyp.value() != 0) hyp.saveTo(counter, out);
        Announce.progressAt(counter++);
      }
      out.close();
    } catch (IOException ex) {
      Announce.failed();
      throw ex;
    }
    Announce.progressDone();
  }

  // -------------------------------------------------------------------------
  //                       Generation
  // -------------------------------------------------------------------------

  /** Computes the measures for a hypothesis */
  public static Measures computeMeasures(Hypothesis hyp) throws SQLException {
    if(hyp.trace()) Announce.message("----------------- Compute",hyp);
    Measures measures = new Measures();
    for (Rule r : Config.rules) {
      r.updateMeasures(hyp, measures);
    }
    hyp.setPriority(measures.priority());
    if(hyp.trace()) Announce.message("Result for",hyp,":",measures.isDetermined()?"determined":"heuristic",measures.truthValue());
    return (measures);
  }

  /** Generate hypotheses, return TRUE if any */
  public static boolean generateHypotheses() throws SQLException {
    hypotheses=new TreeSet<Hypothesis>();
    // Make a nice progress bar
    Integer numPatterns;
    if(Config.delimitPOC==null) {
      numPatterns= Config.database.queryValue("SELECT count(*) FROM patterns", ResultIterator.IntegerWrapper);
    } else {
      numPatterns= Config.database.queryValue("SELECT count(*) FROM patterns, patternOccurrences "+
          "WHERE " + Config.delimitPOC+ " AND patterns.patternid=patternOccurrences.patternid", ResultIterator.IntegerWrapper);
    }
    if (numPatterns == null || numPatterns == 0) return(false);
    Announce.progressStart("Generating all hypotheses", numPatterns);
    isGenerating=true;
    
    // Run through all pattern occurrences
    for (Hypothesis patternOccurrence : new Literal(true, Hypothesis.PATTERNOCCURRENCES, "?ID", "?X", "?Y").instances()) {
      // Run through all associated disambiguation priors
      for (int i = 2; i < 4; i++) {
        if (!Hypothesis.isPolysem(patternOccurrence.component(i))) continue;
        for (Hypothesis disambiguationPrior : new Literal(true, Hypothesis.DISAMBIGUATIONPRIOR, patternOccurrence.component(i), "?X").instances()) {
          computeMeasures(disambiguationPrior);
        }
      }
      computeMeasures(patternOccurrence);
      Announce.progressStep();
    }
    Announce.progressDone();
    Announce.message("Generated",numHypotheses(),"hypotheses");
    isGenerating=false;
    return(hypotheses.size()>0);
  }

  /** Test method*/
  public static void main(String[] args) throws Exception {
    Config.init(new File("sofie.ini"));
    generateHypotheses();
    Announce.debug("Read hypotheses", hypotheses);
  }
}
