package sofie.maxsat;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.filehandlers.FileLines;
import sofie.maxsat.Hypothesis.Truth;

/**
 * Class Rule
 *
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *      
 * This class represents a rule, i.e. a disjunction of literals.
 * 
 * @author Fabian M. Suchanek
 */
public class Rule {

  // ----------------------------------------------------------------------------
  //                    Definition of a rule
  // ----------------------------------------------------------------------------

  /** Counts all existing rules */
  protected static int allIds = 0;

  /** Holds the id of the current rule*/
  protected int id = allIds++;

  /** Holds a description*/
  protected String description;

  /** Holds the literals*/
  protected List<Literal> literals;

  /** Holds the number of free literals*/
  protected int numFreeLiterals = 0;

  /** Holds the weight of this rule*/
  protected double weight;

  /** Given the index of one literal, lists the literals to be instantiated next*/
  protected List<Literal>[] resolveOrder;

  /** Returns the ith literal*/
  public Literal literal(int i) {
    return (literals.get(i));
  }

  /** Returns the Literal from the list that has to be instantiated next, given a set of known variables*/
  protected static Literal resolveNext(List<Literal> toDo, Set<String> knownVariables) {
    Literal best = null;
    int bestValue = -10000;
    for (Literal f : toDo) {
      int value = f.groundPriority(knownVariables);
      if (value > bestValue) {
        bestValue = value;
        best = f;
      }
    }
    toDo.remove(best);
    for (String var : best.components()) {
      if (Literal.isVariable(var)) knownVariables.add(var);
    }
    return (best);
  }

  /** Constructs a rule */
  public Rule(String description, double weight, Literal... literals) {
    this(description, weight, Arrays.asList(literals));
  }

  /** Constructs a rule */
  @SuppressWarnings("unchecked")
  public Rule(String description, double weight, List<Literal> literals) {
    this.description = description;
    this.weight = weight;
    this.literals = literals;
    resolveOrder = new List[literals.size()];
    for (int i = 0; i < literals.size(); i++) {
      // Free literals are only those that we can set to TRUE or FALSE
      if (literal(i).isFree()) numFreeLiterals++;
      List<Literal> toDo = new ArrayList<Literal>(literals);
      Set<String> knownVariables = new TreeSet<String>();
      for (String s : literal(i).components()) {
        if (Literal.isVariable(s)) knownVariables.add(s);
      }
      toDo.remove(literal(i));
      resolveOrder[i] = new ArrayList<Literal>(literals.size());
      for (int j = 0; j < literals.size() - 1; j++) {
        resolveOrder[i].add(resolveNext(toDo, knownVariables));
      }
    }
  }

  @Override
  public String toString() {
    return description == null ? literals.toString() : description + "\n" + literals;
  }

  @Override
  public boolean equals(Object obj) {
    return (obj instanceof Rule && ((Rule) obj).literals.equals(literals));
  }

  @Override
  public int hashCode() {
    return (literals.hashCode());
  }

  public int numLiterals() {
    return (literals.size());
  }

  /** Returns the index of the literal (for some instantiuation of the variables)*/
  public int indexOf(Literal lit) {
    for (int i = 0; i < literals.size(); i++) {
      if (lit.matches(literals.get(i))) {
        return (i);
      }
    }
    return (-1);
  }

  /** returns the rule weight*/
  public double weight() {
    return (weight);
  }

  // ----------------------------------------------------------------------------
  //                    Parsing
  // ----------------------------------------------------------------------------

  /** Parses a line into a rule. */
  public static Rule parse(String line, String description) {
    String originalLine = line;
    List<Literal> literals = new ArrayList<Literal>();
    while (!line.startsWith("=>")) {
      line = line.trim();
      if (line.startsWith("&") || line.startsWith(",")) line = line.substring(1);
      if (line.indexOf(')') == -1) Announce.error("Malformed rule:",originalLine);
      String literalString = line.substring(0, line.indexOf(')') + 1);
      line = line.substring(literalString.length()).trim();
      literals.add(Literal.parse(literalString).swapPolarity());
    }
    line = line.substring(2);
    if (line.contains(")")) {
      literals.add(Literal.parse(line.substring(0, line.indexOf(')') + 1)));
      line = line.substring(line.indexOf(')')+1);
    }
    double weight = 0;
    try {
      weight = Double.parseDouble(line.trim());
    } catch (Exception e) {
      Announce.error("Cannot parse rule weight:", originalLine);
    }
    return (new Rule(description, weight, literals));
  }

  /** Returns the rules defined in that file */
  public static List<Rule> loadFrom(File file) throws IOException {
    Literal FLAGLITERAL = new Literal(true, Hypothesis.FLAGSET, "?X");
    List<Rule> rules = new ArrayList<Rule>();
    String description = null;
    lines: for (String line : new FileLines(file)) {
      // Register a comment as a description
      if (line.startsWith("//") || line.startsWith("#") || line.startsWith("%")) {
        description = line;
        continue;
      }
      // Ignore blank lines
      if (line.trim().length() < 2) continue;
      // Construct the rule, remove flagSet() literals
      Rule rule = parse(line, description);
      for (int i = rule.indexOf(FLAGLITERAL); i != -1; i = rule.indexOf(FLAGLITERAL)) {
        String flag = rule.literal(i).component(1);
        if (!Parameters.isDefined(flag)) Announce.error("Undefined flag:", flag);
        if (rule.literal(i).polarity() == Parameters.getBoolean(flag)) continue lines;
        rule.literals.remove(i);
      }
      rules.add(rule);
    }
    return (rules);
  }

  // ----------------------------------------------------------------------------
  //                    Grounding
  // ----------------------------------------------------------------------------

  /** Adds the measures of unsatisfied rules */
  public void updateMeasures(Hypothesis hyp, Measures measures) throws SQLException {
    for (int startIndex = 0; startIndex < literals.size(); startIndex++) {
      Map<String, String> bindings = literal(startIndex).bindingsToMatch(hyp, null);
      if (bindings != null) {
        if (literal(startIndex).polarityOf(hyp) == Truth.TRUE) continue;
        List<Hypothesis> groundedLiterals = new ArrayList<Hypothesis>(literals.size());
        groundedLiterals.add(hyp);
        if(hyp.trace()) Announce.message("--- Rule",this.literals,"at",startIndex);
        ground(resolveOrder[startIndex], 0, weight(), bindings, literal(startIndex), measures, groundedLiterals);
      }
    }
  }

  /** Helper for groundedrules(), fills a list of grounded rules*/
  protected void ground(List<Literal> literalsToDo, int numUnsatisfied, double weight, Map<String, String> bindings, Literal freeLiteral,
      Measures measures, List<Hypothesis> grounded) throws SQLException {
    boolean trace=grounded.get(0).trace();
    if (literalsToDo.size() == 0) {
      // Come here for all ground clauses that contain the free literal
      // and that are not yet satisfied  
      HypothesisStore.add(grounded);
      measures.addMeasures(numFreeLiterals, numUnsatisfied, freeLiteral.polarity(), weight);
      measures.addRelatedHypotheses(grounded);
      if(trace) {
        Announce.message(Announce.blanks(grounded.size()),"Rule is unhappy with the following groundings\n",grounded); 
        Announce.message(Announce.blanks(grounded.size()),measures);
      }      
      return;
    }
    // Prepare the recursion
    Literal currentLiteral = literalsToDo.get(0).withBindings(bindings);
    literalsToDo = literalsToDo.subList(1, literalsToDo.size());
    grounded.add(null);
    if(trace) Announce.message(Announce.blanks(grounded.size()),"Grounding",currentLiteral);
    int newNumUnsatisfied=numUnsatisfied+(currentLiteral.isFree()?1:0);
    boolean foundGrounding=false;
    // Search only those groundings that are unhappy or unknown
    for (Hypothesis f : currentLiteral.ground()) {
      foundGrounding=true;
      if(trace) Announce.message(Announce.blanks(grounded.size()),"Checking ground instance",f);
      grounded.set(grounded.size() - 1,f);
      if(f.trace()) Announce.message(Announce.blanks(grounded.size()),"--- Rule",this.literals,grounded);
      // For Non-YAGO-Relations, ignore the literal and pass on the weight
      if (f.relationIfKnown()==Hypothesis.DISAMBIGUATIONPRIOR || f.relationIfKnown()==Hypothesis.PATTERNOCCURRENCES) {
        if(trace) Announce.message(Announce.blanks(grounded.size()),f,"has predefined weight");
        ground(literalsToDo, newNumUnsatisfied, weight * f.value(), currentLiteral.bindingsToMatch(f, bindings), freeLiteral, measures, grounded);
        continue;
      }
      // For all other relations, check the value of the literal
      switch (currentLiteral.polarityOf(f)) {
        case FALSE: // We have one more unhappy literal
          if(trace) Announce.message(Announce.blanks(grounded.size()),"makes the rule unsatisfied");
          ground(literalsToDo, newNumUnsatisfied, weight, currentLiteral.bindingsToMatch(f, bindings), freeLiteral, measures, grounded);
          continue;
        case UNKNOWN: // We have an literal with an unknown value        
          if(trace) Announce.message(Announce.blanks(grounded.size()),"has an unknown value");
          ground(literalsToDo, numUnsatisfied, weight, currentLiteral.bindingsToMatch(f, bindings), freeLiteral, measures, grounded);
          continue;
        case TRUE: // Any positive value => This rule is already happy
          // This should be filtered out automatically by Literal.ground()
          if(trace) Announce.message(Announce.blanks(grounded.size()),"satisfies the literal and is ignored");
          continue;
      }
    }
    if(!foundGrounding && trace) Announce.message(Announce.blanks(grounded.size()),"No ground instance found");
    grounded.remove(grounded.size() - 1);
  }

  /** Test */
  public static void main(String[] args) throws Exception {
    Parameters.init("sofie.ini");
    D.p(loadFrom(new File("./sofie/config/rules.txt")));
  }
}