package sofie.parsing;

import java.util.regex.Pattern;

import javatools.administrative.D;
import javatools.parsers.Name;
import basics.datagraph.LiteralYagoClass;
import basics.datagraph.YagoClasses;

/**
 * Class Token
 * 
 * This class is part of the SOFIE system (http://mpii.de/yago-naga/sofie). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share-Alike 3.0 Unported License 
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) by Fabian M. Suchanek (http://suchanek.name).
 * If you use this class for scientific purposes, please cite
 *      Fabian M. Suchanek, Mauro Sozio, Gerhard Weikum
 *      "SOFIE: A Self-Organizing Framework for Information Extraction"
 *      (International World Wide Web Conference 2009)
 *
 * This class represents a token in a natural language document.
 * 
 * @author Fabian M. Suchanek
 */
public class Token implements Comparable<Token> {

  //---------------------------------------------------------------------------
  //                  Default Tokens
  //---------------------------------------------------------------------------

  /** Holds the token itself*/
  protected String token;

  /** Holds the original word */
  protected String original = null;

  /** Constructs a token*/
  public Token(String s) {
    token = original = s.intern();
  }

  /** Tells whether the token shall be considered as a possible entity*/
  public boolean isInteresting() {
    return (false);
  }

  /** Tells whether the token shall be replaced by '@' in a pattern*/
  public boolean replaceMe() {
    return (false);
  }

  /** Returns the token itself*/
  public String token() {
    return (token);
  }

  /** Returns the orignal word*/
  public String original() {
    return original;
  }

  /** TRUE for words containing '_' or ' '*/
  public boolean isMultiWord() {
    if(token.length()<4) return(false);
    String s=token.substring(1,token.length()-1);
    return(s.contains(" ") || s.contains("_"));
  }
  
  /** Compares by name and type*/
  public int compareTo(Token o) {
    return D.comparePairs(this.getClass().getName(), o.getClass().getName(), token(), o.token());
  }

  @Override
  public boolean equals(Object obj) {
    return obj != null && this.getClass().isInstance(obj) && token.equals(((Token) obj).token);
  }

  @Override
  public int hashCode() {
    return this.getClass().getName().hashCode() ^ token().hashCode();
  }

  @Override
  public String toString() {
    return this.getClass().getSimpleName() + ": " + token;
  }

  //---------------------------------------------------------------------------
  //                  Special types of tokens
  //---------------------------------------------------------------------------

  // ---------------------- Word and punctuation tokens ----------------------------------

  /** Normal word tokens */
  public static class Word extends Token {

    public Word(String s) {
      super(s);
    }
  }

  /** Punctuation tokens */
  public static class Punctuation extends Token {

    public Punctuation(String s) {
      super(s);
    }
  }

  /** Stop word tokens */
  public static class StopWord extends Word {

    public StopWord(String s) {
      super(s);
    }
  }

  //---------------------- Literal tokens ----------------------------------

  /** Literal tokens */
  public static abstract class Literal extends Token {

    /** Point to the literal YAGO class*/
    public abstract LiteralYagoClass yagoClass();

    public Literal(String s) {
      super(s);
    }

    @Override
    public boolean replaceMe() {
      return (true);
    }
  }

  /** Normalized dates */
  public static class Date extends Literal {

    public Date(String s) {
      super(s.matches("\\d{4}") ? s + "-##-##" : s);
    }

    @Override
    public boolean isInteresting() {
      return (true);
    }

    @Override
    public LiteralYagoClass yagoClass() {
      return (YagoClasses.DATE);
    }
  }

  /** Numbers */
  public static class Number extends Literal {

    public Number(String s) {
      super(s);
    }

    public LiteralYagoClass yagoClass() {
      return (YagoClasses.NUMBER);
    }

    @Override
    public boolean isInteresting() {
      return token.contains("#") || token.length() > 3;
    }
  }

  /** URLs */
  public static class URL extends Literal {

    public URL(String s) {
      super(s);
    }

    public LiteralYagoClass yagoClass() {
      return (YagoClasses.URL);
    }
  }

  //---------------------- Common Individuals ----------------------------------

  /** Types of a common name*/
  public enum NameType {
    /** Canonic, i.e. identical to the title of a Wikipedia article and known to be so*/
    CANONIC,
    /** Polysemous, i.e. a WIC and known to be so*/
    POLYSEM,
    /** Without any meaning and known to be so*/
    VOID,
    /** Not yet established type*/
    UNKNOWN,
    /** Wikipedia link, but not necessarily yet the exact entity name*/
    SEMICANONIC
  };

  /** Superclass for common individuals*/
  public static class CommonIndividual extends Token {

    /** Type of the token*/
    protected NameType nameType = NameType.UNKNOWN;

    public CommonIndividual(String s) {
      super(s);
    }

    /** Returns the type */
    public NameType nameType() {
      return (nameType);
    }

    @Override
    public boolean replaceMe() {
      return (true);
    }

    /** Marks this token as disambiguated*/
    public void disambiguateAs(String n) {
      original = token;
      token = n;
      nameType = NameType.CANONIC;
    }

    /** Marks this token as polysemous*/
    public void makePolysem(int docId) {
      original = token;
      token = '^' + token + docId;
      nameType = NameType.POLYSEM;
    }

    /** Marks this token as meaningless*/
    public void makeVoid() {
      nameType = NameType.VOID;
    }

    @Override
    public String toString() {
      return token+"/"+nameType;
    }
    
    /** Redirects this token to the given token*/
    public void redirectTo(CommonIndividual t) {
      this.token=t.token;
      this.nameType=t.nameType;
    }
  }

  /** Named entities*/
  public static class ProperName extends CommonIndividual {

    public ProperName(String s) {
      super(s);
    }

    @Override
    public boolean isInteresting() {
      return true;
    }
  }

  /** Semi-Disambiguated proper names (like wikipedia links)*/
  public static class SemiCanonicProperName extends ProperName {

    public SemiCanonicProperName(String s) {
      super(s);
      // If this is a Wikipedia link in the wiki markup language, take directly the link
      String[] split = s.split("\\s*\\|\\s*");
      if (split.length == 2) {
        if (Name.isLanguage(split[1].trim())) {
          token = split[1].trim() + "_language";
          nameType = NameType.CANONIC;
          return;
        }
        token = split[0].trim();
      }
      nameType = NameType.SEMICANONIC;
    }

  }

  /** US States */
  public static class USState extends SemiCanonicProperName {

    public USState(String s) {
      super(s);
    }
  }

  /** Disambiguated proper names*/
  public static class CanonicProperName extends ProperName {

    public CanonicProperName(String s) {
      super(s);
      nameType = NameType.CANONIC;
    }

  }

  /** Persons*/
  public static class Person extends ProperName {

    /** Holds the name*/
    protected Name.PersonName personName;

    public Person(String s) {
      this(new Name.PersonName(s));
    }

    public Person(Name.PersonName n) {
      super(n.normalize());
      personName = n;
    }

    public Name.PersonName personName() {
      return (personName);
    }

    @Override
    public String toString() {
      return "Person: " + token;
    }
  }

  /** Companies*/
  public static class Company extends ProperName {

    public Company(String s) {
      super(s);
    }

    @Override
    public String toString() {
      return super.toString();
    }
  }

  /** non-proper-name Wikipedia links*/
  public static class WikiLink extends CommonIndividual {

    public WikiLink(String s) {
      super(s);
      nameType = NameType.SEMICANONIC;
    }
  }

  /** Superclass for tokens that are considered in patterns*/
  public static class AnyName extends Token {

    public AnyName(String s) {
      super(s);
    }

    @Override
    public boolean replaceMe() {
      return (true);
    }
  }

  // --------------- Function tokens ---------------------------------

  /** Tokens that serve purely functional purposes */
  public static class Functional extends Token {

    public Functional(String s) {
      super(s);
    }
  }

  /** Tokens that cannot be part of a pattern*/
  public static class Separator extends Functional {

    public Separator(String s) {
      super(s);
    }
  }

  /** Generic STOP token*/
  public static final Separator SEPARATOR = new Separator("STOP");

  /** Token that starts a comment (e.g. <!-- --> in HTML) */
  public static class StartCommentToken extends Functional {

    /** String that ends the comment*/
    public final Pattern end;

    /** Constructs a comment token with a case insensitive pattern as end*/
    public StartCommentToken(String commentType, String end) {
      super(commentType);
      this.end = Pattern.compile(end, Pattern.CASE_INSENSITIVE);
    }
  }

  /** HTML script start*/
  public static final StartCommentToken STARTSCRIPT = new StartCommentToken("SCRIPT", "</SCRIPT>");

  /** HTML comment start*/
  public static final StartCommentToken HTMLCOMMENT = new StartCommentToken("COMMENT", "-->");

  /** Wikipedia special boxes*/
  public static final StartCommentToken STARTBRACES = new StartCommentToken("{{...}}", "\\}\\}");

  /** Wikipedia Infobox*/
  public static final StartCommentToken STARTINFOBOX = new StartCommentToken("{{Infobox...}}", "'''");

  /** Wikipedia Article*/
  public static final StartCommentToken STARTARTICLE = new StartCommentToken("Article", "DontFindThisString");

  /** Wikipedia revisions*/
  public static final StartCommentToken REVISION = new StartCommentToken("REVISION", "<text xml:space=\"preserve\">");

  /** Wikipedia references*/
  public static final StartCommentToken REF = new StartCommentToken("REFERENCE", "</ref>|/>");

  /** Token that can be ignored*/
  public static final Token IGNORE = new Token("IGNORE");

  /** Token that shall be repeated before repeat tokens*/
  public static class Title extends Functional {

    public Title(String s) {
      super(s);
    }
  }

  /** A token that places the title token left of itself*/
  public static class Repeat extends Functional {

    public Repeat(String s) {
      super(s);
    }
  }

  /** Head of a Wikipedia infobox. 
   * Is a repeat token, so that the infobox type can become a pattern*/
  public static final Repeat INFOBOXHEAD = new Repeat("INFOHEAD");

  /** A new line of a Wikipedia infobox*/
  public static final Repeat INFOBOXBAR = new Repeat("INFO");

  /** A Wikipedia category*/
  public static final Repeat CATEGORY = new Repeat("CAT");
}